"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit curtain walls utility functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    BuiltInCategory,
    FilteredElementCollector,
    WallKind,
)

from duHast.Revit.Walls.Utility.walls_type_sorting import sort_wall_types_by_family_name

#: Built in wall family name for curtain wall
CURTAIN_WALL_FAMILY_NAME = WallKind.Curtain


def get_all_curtain_wall_type_ids(doc):
    """
    Gets type ids off all available curtain wall types in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing curtain wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    dic = sort_wall_types_by_family_name(doc)
    if CURTAIN_WALL_FAMILY_NAME in dic:
        ids = dic[CURTAIN_WALL_FAMILY_NAME]
    return ids


def get_all_curtain_wall_instances(doc, available_ids):
    """
    Gets all curtain wall elements placed in model...ignores legend elements.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param available_ids: Filter: curtain wall type ids to check wall instances for.
    :type available_ids: list of Autodesk.Revit.DB.ElementId
    :return: List of wall instances
    :rtype: List of Autodesk.Revit.DB.Wall
    """

    instances = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Walls)
        .WhereElementIsNotElementType()
    )
    for c in col:
        if c.GetTypeId() in available_ids:
            instances.append(c)
    return instances


def get_placed_curtain_wall_type_ids(doc, available_ids):
    """
    Gets all used curtain wall types in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param available_ids: Filter: curtain wall type ids to check wall types for.
    :type available_ids: list of Autodesk.Revit.DB.ElementId
    :return: List of wall instances
    :rtype: List of Autodesk.Revit.DB.Wall
    """

    instances = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Walls)
        .WhereElementIsNotElementType()
    )
    for c in col:
        if c.GetTypeId() in available_ids:
            instances.append(c.GetTypeId())
    return instances



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to retrieving file information. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import glob

from duHast.Utilities.files_io import get_file_name_without_ext
import os


def get_files_single_directory(folder_path, file_prefix, file_suffix, file_extension):
    """
    Get files from a folder filtered by file prefix, file suffix, file extension
    :param folder_path: Folder path from which to get files.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value.
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :return: A list of all the files matching the supplied filters.
    :rtype: list str
    """

    fileList = glob.glob(
        os.path.join(folder_path, file_prefix + "*" + file_suffix + file_extension)
    )
    return fileList


def get_files_from_directory_walker_with_filters(
    folder_path, file_prefix, file_suffix, file_extension
):
    """
    Returns a list of all files in directory and nested sub directories where file name matches filters value.
    :param folder_path: Root folder path from which to get files.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :return: A list of all the files matching the supplied filters.
    :rtype: list str
    """

    files_found = []
    for root, dirs, files in os.walk(folder_path):
        for name in files:
            file_name = get_file_name_without_ext(name)
            if (
                name.endswith(file_extension)
                and file_name.startswith(file_prefix)
                and file_name.endswith(file_suffix)
            ):
                files_found.append(os.path.join(root, name))
    return files_found


def get_files_from_directory_walker_with_filters_simple(folder_path, file_extension):
    """
    Returns a list of all files in directory and nested subdirectories where file name matches file extension filter value
    :param folder_path: Root folder path from which to get files.
    :type folder_path: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :return: A list of all the files matching the supplied filters.
    :rtype: list str
    """

    files_found = []
    files_found = get_files_from_directory_walker_with_filters(
        folder_path, "", "", file_extension
    )
    return files_found


def files_as_dictionary(
    folder_path, file_prefix, file_suffix, file_extension, include_sub_dirs=False
):
    """
    Returns a dictionary of all files in directory and nested subdirectories where file name contains filter value.
    - key file name without extension
    - values: list of directories where this file occurs (based on file name only!)
    Use case: check for duplicates by file name only

    :param folder_path: Root folder path from which to get files.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param include_sub_dirs: If True subdirectories will be included in search for files, defaults to False
    :type include_sub_dirs: bool, optional
    :return: A dictionary where the key is the file name without the file extension. Value is a list of fully qualified file path to instances of that file.
    :rtype: dictionary
        key: str
        value: lit of str
    """

    files_found = []
    # set up a dictionary
    file_dic = {}
    try:
        if include_sub_dirs:
            files_found = get_files_from_directory_walker_with_filters(
                folder_path, file_prefix, file_suffix, file_extension
            )
        else:
            files_found = get_files_single_directory(
                folder_path, file_prefix, file_suffix, file_extension
            )
    except Exception:
        return file_dic

    # populate dictionary
    for file_path in files_found:
        file_name = get_file_name_without_ext(file_path)
        if file_name in file_dic:
            file_dic[file_name].append(file_path)
        else:
            file_dic[file_name] = [file_path]
    return file_dic


def get_files(folder_path, file_extension=".rvt"):
    """
    Gets a list of files from a given folder with a given file extension.
    (Returns the full path for each file!)

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_extension: Filter: File needs to have this file extension, defaults to '.rvt'
    :type file_extension: str, optional
    :return: List of file path
    :rtype: list of str
    """

    file_list = glob.glob(os.path.join(folder_path, "*" + file_extension))
    return file_list


def get_files_with_filter(folder_path, file_extension=".rvt", filter="*"):
    """
    Gets a list of files from a given folder with a given file extension and a matching a file name filter.
    (Returns the full path for each file!)

    :param folder_path: Folder path from which to get files.
    :type folder_path: str
    :param file_extension: Filter: File needs to have this file extension, defaults to '.rvt'
    :type file_extension: str, optional
    :param filter: File name filter ('something*'), defaults to '*'
    :type filter: str, optional
    :return: List of file path
    :rtype: list of str
    """

    file_list = glob.glob(os.path.join(folder_path, filter + file_extension))
    return file_list


def get_files_from_directory_walker(path, filter):
    """
    Gets all files in directory and nested subdirectories where file name contains filter value.

    :param path: Folder path from which to get files.
    :type path: str
    :param filter: File name filter ('something')
    :type filter: str
    :return: List of file path
    :rtype: list of str
    """

    # file_list = glob.glob(os.path.join(path, "*" + filter + "*.*"), recursive=True)
    file_list = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if filter in name:
                file_list.append(os.path.join(root, name))
    return file_list





"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class representing the shared parameter model.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import json

from duHast.Utilities.Objects.base import Base

class ParameterModel(Base):

    """
    A class representing the shared parameter model.
    """
    def __init__(self, name = None, group=None, is_type_parameter = True, para_type=None, visiblity=False, property_group=None, shared_parameter_file_path=None, j=None):
        """
        Constructor for the parameter data class.

        :param name: The name of the parameter.
        :type name: str
        :param group: The group of the parameter. (within the sharted parameter file)
        :type group: str
        :param is_type_parameter: Is the parameter a type parameter (True) or instance parameter (False)
        :type is_type_parameter: bool
        :param para_type: The type of the parameter. (Text, Number, etc.)
        :type para_type: str
        :param visiblity: The visiblity of the parameter.
        :type visiblity: bool
        :param property_group: The property group of the parameter. (where it will appear in revit UI under)
        :type property_group: str
        :param shared_parameter_file_path: The path to the shared parameter file.
        :type shared_parameter_file_path: str

        """

        super(ParameterModel, self).__init__()

        self._name = name
        self._group = group
        self._is_type_parameter = is_type_parameter
        self._para_type = para_type
        self._visibility = visiblity
        self._property_group = property_group
        self._shared_parameter_file_path = shared_parameter_file_path

        json_var = None
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # no action required
                json_var = j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )
        
            try:
                self._name = (json_var.get("name",None,))
                self._group = (json_var.get("group",None,))
                self._para_type = (json_var.get("para_type",None,))
                self._visibility = (json_var.get("visibility",None,))
                self._property_group = (json_var.get("property_group",None,))
                self._shared_parameter_file_path = (json_var.get("shared_parameter_file_path",None,))
            
            except Exception as e:
                raise type(e)(
                    "Node {} failed to initialise with: {}".format(self.data_type, e)
                )
        
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        self._name = value
    
    @property
    def group(self):
        return self._group

    @group.setter
    def group(self, value):
        self._group = value

    @property
    def is_type_parameter(self):
        return self._is_type_parameter
    
    @is_type_parameter.setter
    def is_type_parameter(self, value):
        self._is_type_parameter = value
    
    @property
    def para_type(self):
        return self._para_type

    @group.setter
    def para_type(self, value):
        self._para_type = value
    
    @property
    def visibility(self):
        return self._visibility
    
    @visibility.setter
    def visibility(self, value):
        self._visibility = value
    
    @property
    def property_group(self):
        return self._property_group
    
    @property_group.setter
    def property_group(self, value):
        self._property_group = value
    
    @property
    def shared_parameter_file_path(self):
        return self._shared_parameter_file_path
    
    @shared_parameter_file_path.setter
    def shared_parameter_file_path(self, value):
        self._shared_parameter_file_path = value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around creating Revit filled regions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction


# import Autodesk
from Autodesk.Revit.DB import (
    FilledRegion,
    Transaction,
)

def create_filled_region_by_view(doc, view, curve_loops, filled_region_type, transaction_manager = in_transaction):
    """
    Creates a filled region in the specified view.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to create the filled region in.
    :type view: Autodesk.Revit.DB.View
    :param curve_loops: The curve loops to use for the filled region.
    :type curve_loops: list Autodesk.Revit.DB.CurveLoop
    :param filled_region_type: The type of filled region to create.
    :type filled_region_type: Autodesk.Revit.DB.ElementId

    :return: A result object containing the created filled region instance.
    :rtype: Result
    """


    # set up a status tracker
    return_value = Result()

    # set up an action creating a filled region
    def action():
         # set up a status tracker
        action_return_value = Result()
        try:
            # create the filled region
            filled_region = FilledRegion.Create(doc, filled_region_type, view.Id, curve_loops)
            # store the filled region in the return value
            action_return_value.result.append(filled_region)
            action_return_value.append_message("Filled region created successfully.")
        except Exception as e:
            action_return_value.update_sep(False, "Failed to create filled region with error: {}".format(e))
        return action_return_value

    if transaction_manager:
        transaction = Transaction(doc, "Drawing filled region")
        return_value = transaction_manager(transaction, action)
    else:
        return_value = action()

    return return_value


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from Autodesk.Revit.DB import  ColorDepthType, ExportPaperFormat, PDFExportOptions, PDFExportQualityType


def set_pdf_export_option_2024(
        naming_rule, 
        always_use_raster = False, 
        color_depth = ColorDepthType.Color,
        combine = False,
        paper_format =  ExportPaperFormat.Default,
        export_quality = PDFExportQualityType.DPI600,
        hide_crop_boundaries = True,
        hide_scope_boxes = True,
        hide_unreferenced_view_tags = True,
        mask_coincident_lines = True,
        replace_halftone_with_thin_lines = False,
        stop_on_error = False,
        view_links_in_blue = False,
        ):
    """
    Sets the PDF export options for the Revit document.
    :param naming_rule: The naming rule for the PDF export.
    :type naming_rule: str
    
    :return: The PDF export options.
    :rtype: Autodesk.Revit.DB.PDFExportOptions
    """

    pdf_export_option = PDFExportOptions()

    # Set the naming rule if valid
    if (PDFExportOptions.IsValidNamingRule(naming_rule)):
        # Set the naming rule for the PDF export options
        pdf_export_option.SetNamingRule(naming_rule)
    else:
        # If the naming rule is not valid, set it to None
        pdf_export_option.SetNamingRule(None)
        return None

    pdf_export_option.AlwaysUseRaster = always_use_raster
    pdf_export_option.ColorDepth =  color_depth
    pdf_export_option.Combine = combine
    pdf_export_option.PaperFormat = paper_format # use sheet size
    pdf_export_option.ExportQuality = export_quality

    pdf_export_option.HideCropBoundaries = hide_crop_boundaries
    pdf_export_option.HideScopeBoxes = hide_scope_boxes
    pdf_export_option.HideUnreferencedViewTags = hide_unreferenced_view_tags
    pdf_export_option.MaskCoincidentLines = mask_coincident_lines
    pdf_export_option.ReplaceHalftoneWithThinLines = replace_halftone_with_thin_lines
    pdf_export_option.StopOnError = stop_on_error
    pdf_export_option.ViewLinksInBlue = view_links_in_blue

    return pdf_export_option


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing comparison reporting of family types in a project vs in a library functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reports:

- if family does not exist in library
- if type does not exist in library
- if a parameter value for a given type is different to the parameter value for that type in the library

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os

from Autodesk.Revit.DB import Family

from duHast.Revit.Family.family_types_get_data_from_xml import get_family_type_data_from_library
from duHast.Revit.Family.family_types_model_get_data_from_xml import get_type_data_via_XML_from_family_object
from duHast.Revit.Family.family_functions import get_name_to_family_dict
from duHast.Utilities.files_xml import get_all_xml_files_from_directories
from duHast.Utilities.files_csv import read_csv_file
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects.result import Result
from duHast.UI.Objects.ProgressBase import ProgressBase
from duHast.Revit.Family.Data.Objects.family_type_data_storage_manager import (
    FamilyTypeDataStorageManager,
)


def get_family_type_data_from_project_file(
    doc, type_data_from_library, progress_callback=None
):
    """
    Get the family type data from the families in the project file of families that are in the library only.

    :param doc: Revit document
    :type doc: Autodesk.Revit.DB.Document
    :param type_data_from_library: list of family type data from the library
    :type type_data_from_library: [:class:`FamilyTypeDataStorageManager`]
    :param progress_callback: progress callback object
    :type progress_callback: :class:`ProgressBase`

    :return:
        Result class instance.

        - result.status: XML conversion status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain which xml file was read and converted into family type data.
        - result.result will be [:class:`FamilyTypeDataStorageManager`]

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = Result()
    matched_data = []
    try:

        # get all families in the project
        families_loaded = get_name_to_family_dict(doc)

        # set progress counter
        counter = 1
        max_value_xml = len(families_loaded)

        # loop over loaded families and search for matches based on name and category
        for fam_name, revit_family in families_loaded.items():

            # update progress
            if progress_callback:
                progress_callback.update(counter, max_value_xml,fam_name)

            # check this is a family
            if isinstance(revit_family, Family) is False:
                return_value.append_message(
                    "skipping family: {} as it is not a family".format(fam_name)
                )
                continue

            # ignore in place families
            if revit_family.IsInPlace:
                continue

            # get the family category
            fam_cat = revit_family.FamilyCategory.Name

            # get the family type data storage from the library
            found_match = False
            match_library = None
            for type_data_storage_manager_library in type_data_from_library:
                if (
                    type_data_storage_manager_library.family_type_data_storage[
                        0
                    ].family_name
                    == fam_name
                    and type_data_storage_manager_library.family_type_data_storage[
                        0
                    ].root_category_path
                    == fam_cat
                ):
                    # found a match...
                    found_match = True
                    match_library = type_data_storage_manager_library
                    break

            # check if match in library was found
            if found_match is False:
                # no match found
                matched_data.append(([fam_name, fam_cat], None))
                # update progress
                counter = counter + 1
                continue

            # create temp xml files from loaded family
            type_data_result = get_type_data_via_XML_from_family_object(
                revit_family=revit_family
            )
            if type_data_result.status == False:
                return_value.update_sep(
                    False,
                    "Failed to get type data from family: {} with exception: {}".format(
                        fam_name, type_data_result.message
                    ),
                )
                matched_data.append((type_data_storage_manager_loaded_fam, None))
                # update progress
                counter = counter + 1
                continue

            # get the type data of the family
            type_data_storage_manager_loaded_fam = type_data_result.result[0]
            # add to matched data
            matched_data.append((type_data_storage_manager_loaded_fam, match_library))

            # update progress
            counter = counter + 1

            # check for user cancel
            if progress_callback != None:
                if progress_callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )

    # store data to be returned
    return_value.result = matched_data

    return return_value


def build_comparison_report(type_data_matches, ignore_list_path):
    """
    Build a comparison report of the family type data from the project file against the library.
    Only differences are reported.

    :param type_data_matches: list of matched family type data
    :type type_data_matches: [([:class:`FamilyTypeDataStorageManager`], [:class:`FamilyTypeDataStorageManager`])]
    :param ignore_list_path: path to a csv file containing a list of families to ignore (name, category) in the comparison report.
    :type ignore_list_path: str

    :return:
        Result class instance.

        - result.status: Comparison status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message: will be empty.
        - result.result will be [[str]] where each entry is a list of family name, category etc and difference.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    # families not found in library, used to avoid duplicate entries
    not_in_library = []
    # list of differences
    diff = []
    return_value = Result()
    try:

        # get ignore data
        ignore_data = []
        if ignore_list_path != None:
            ignore_data_result = read_csv_file(ignore_list_path)
            if ignore_data_result.status == False:
                return_value.update_sep(False, ignore_data_result.message)
                return return_value
            else:
                ignore_data = ignore_data_result.result

        # loop over the data and compare
        for entry in type_data_matches:
            fam_name = ""
            fam_category = ""
            # set a flag that the family export from revit project file resulted in no types exported
            family_export_has_types = False

            # get name and category ( for non matched family this may just be a list of name and category rather than a storage object)
            if isinstance(entry[0], list):
                fam_name = entry[0][0]
                fam_category = entry[0][1]
            elif isinstance(entry[0], FamilyTypeDataStorageManager):
                fam_name = entry[0].family_name
                fam_category = entry[0].family_category
                # there is a chance that the family export from the project file contains no types
                if entry[0].family_has_types:
                    # set flag that types were exported
                    family_export_has_types = True
            else:
                raise ValueError(
                    "entry[0] is not a list or FamilyTypeDataStorageManager: {}".format(
                        type(entry[0])
                    )
                )

            # check if the family is in ignore list based on name and category
            ignore = False
            for ignore_entry in ignore_data:
                if ignore_entry[0] == fam_name and ignore_entry[1] == fam_category:
                    ignore = True
                    break
            if ignore:
                continue

            if entry[1] == None:
                if "{}{}".format(fam_name, fam_category) not in not_in_library:
                    not_in_library.append("{}{}".format(fam_name, fam_category))
                    # family not found in library
                    diff.append([fam_name, fam_category, "No match in library"])
                continue
            elif family_export_has_types == False:
                # family has no types
                diff.append(
                    [fam_name, fam_category, "Family in project has no types exported"]
                )
            else:
                # compare the two data sets
                diff = diff + entry[0].get_difference(entry[1])

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )

    # store data to be returned
    return_value.result = diff

    return return_value


def compare_family_files_in_project_against_library(
    doc, process_directories, ignore_list_path=None, progress_callback=None
):
    """
    Compare the family type data from the project file against the library.
    Only differences are reported.

    :param doc: Revit document
    :type doc: Autodesk.Revit.DB.Document
    :param process_directories: list of directories to search for xml files
    :type process_directories: list
    :param ignore_list_path: path to a csv file containing a list of families to ignore (name, category) in the comparison report.
    :type ignore_list_path: str
    :param progress_callback: progress callback object
    :type progress_callback: :class:`ProgressBase`

    :return:
        Result class instance.

        - result.status: Comparison status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message: Log entries.
        - result.result will be [[str]] where each entry is a list of family name, category etc and difference.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )

    # set up a timer
    t = Timer()
    t.start()

    try:

        # get all xml files from the directory representing families in the library (point of truth)
        xml_files_in_libraries = get_all_xml_files_from_directories(process_directories)

        # check if any xml files were found
        if len(xml_files_in_libraries) == 0:
            return_value.update_sep(
                False,
                "No XML files found in the directories: {}".format(process_directories),
            )
            return_value.append_message(t.stop())
            return return_value
        
        return_value.append_message(
            "Found {} XML files in the directories: {}".format(
                len(xml_files_in_libraries), process_directories
            )
        )

        # get the type data from the library
        type_data_from_library_result = get_family_type_data_from_library(
            xml_files_in_libraries, progress_callback
        )

        # check if the type data from the library was successfully gathered
        if (
            type_data_from_library_result.status == False
            or len(type_data_from_library_result.result) == 0
        ):
            return_value.update_sep(False, type_data_from_library_result.message)
            return_value.append_message(t.stop())
            return return_value
        
        return_value.append_message(
            "Successfully gathered family type data from the library."
        )

        # get type data from the family files in project file
        type_data_from_project_result = get_family_type_data_from_project_file(
            doc, type_data_from_library_result.result, progress_callback
        )

        # check if the type data from the project file was successfully gathered
        if (
            type_data_from_project_result.status == False
            or len(type_data_from_project_result.result) == 0
        ):
            return_value.update_sep(False, type_data_from_project_result.message)
            return_value.append_message(t.stop())
            return return_value
        
        return_value.append_message(
            "Successfully gathered family type data from the project file."
        )

        # compare the data
        comparison_report_result = build_comparison_report(
            type_data_from_project_result.result, ignore_list_path
        )
        if comparison_report_result.status == False:
            return_value.update_sep(False, comparison_report_result.message)
            return_value.append_message(t.stop())
            return return_value
        
        return_value.append_message(
            "Successfully compared family type data from project file against library."
        )
        return_value.append_message(t.stop())

        # store the comparison report as sorted list by family name
        sorted_list = sorted(comparison_report_result.result, key=lambda x: x[0])

        # add the file name the family is in as the first entry
        for entry in sorted_list:
            entry.insert(0, os.path.basename(doc.Title))

        return_value.result = sorted_list

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )
        if t.is_running():
            return_value.append_message(t.stop())

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of helper functions relating to pyrevit directory selection.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Utilities.directory_io import get_all_nested_directories

# set up some options for the user to select
SINGLE_FOLDER_OPTION = "Single folder"
NESTED_FOLDER_OPTION = "Include nested folders"


def get_user_options(forms):
    """
    Gets the user options for saving out families

    Options are: save out to single directory, exclude existing families, overwrite existing families

    :param forms: the forms object
    :type forms: Forms
    :return: the user options
    :rtype: dict
    """

    # ask user for single directory or directory by category
    # and if existing families are to be ignored
    ops = [SINGLE_FOLDER_OPTION,NESTED_FOLDER_OPTION]
    configs = {
        SINGLE_FOLDER_OPTION: {"background": "0xFF55FF"},
        NESTED_FOLDER_OPTION: {"background": "0xFF55FF"},
    }
    ui_options = forms.CommandSwitchWindow.show(
        ops,  message="Select Directory Option", config=configs
    )

    return ui_options


def get_process_directories(forms, form_title="Select directory"):
    """
    Gets the directories to process

    :param forms: the forms object
    :type forms: Forms
    :return: the directories to process
    :rtype: list
    """

    # set up a status tracker
    return_value = Result()
    try:

        # get the library directory path
        library_path = forms.pick_folder(title=form_title)
        if not library_path:
            message = "No directory selected."
            print(message)
            return_value.update_sep(False, message)
            return return_value
        
        # display UI to get user options re single or nested folders
        user_option =  get_user_options(forms)
        
        # check what came back
        if user_option == SINGLE_FOLDER_OPTION:
            # get the library directory path
            return_value.result.append (library_path)
        elif user_option == NESTED_FOLDER_OPTION:
            # get all subdirs
            child_dirs = get_all_nested_directories(library_path)
            # get all children
            return_value.result=child_dirs
            # make sure root dir is included too
            return_value.result.append(library_path)
        else:
            return_value.update_sep(False, "No valid directory option selected")
    except Exception as e:
        return_value.update_sep(
            False, "Failed to get directories: {}".format(e)
        )

    return return_value


def get_process_directory(forms, form_title="Select directory"):
    """
    Gets the directories to process

    :param forms: the forms object
    :type forms: Forms
    :return: the directories to process
    :rtype: list
    """

    # set up a status tracker
    return_value = Result()
    try:

        # get the library directory path
        library_path = forms.pick_folder(title=form_title)
        if not library_path:
            message = "No directory selected."
            print(message)
            return_value.update_sep(False, message)
            return return_value
        
       
        # get the library directory path
        return_value.result.append (library_path)
        
    except Exception as e:
        return_value.update_sep(
            False, "Failed to get directories: {}".format(e)
        )

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around exporting from Revit to varies file formats.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

#
# doSomethingWithViewName method will accept view name as arg only
def build_export_file_name_from_view(view_name, view_filter_rule, file_extension):
    """
    Function modifying the past in view name and returns a file name.

    :param view_name: The view name to be used as file name.
    :type view_name: str
    :param view_filter_rule: A prefix which will be removed from the view name.
    :type view_filter_rule: str
    :param file_extension: The file extension to be used. Format is '.something'
    :type file_extension: str
    :return: A file name.
    :rtype: str
    """

    # check if file extension is not none
    if file_extension is None:
        file_extension = ".tbc"
    # check the filter rule
    if view_filter_rule is None:
        new_file_name = view_name + file_extension
    else:
        new_file_name = view_name[len(view_filter_rule) :] + file_extension
        new_file_name = new_file_name.strip()
    return new_file_name



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Copies annotation samples.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# required for ToList() call
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from Autodesk.Revit.DB import (
    CopyPasteOptions,
    Document,
    ElementId,
    ElementTransformUtils,
    Transaction,
    Transform,
    View,
)

from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction
from duHast.UI.Objects.ProgressBase import ProgressBase


def copy_annotation_to_views(
    doc,
    source_dims_by_view_name,
    active_view,
    views_by_name_selected,
    callback_progress=None,
):
    """
    Copy annotation from one view to another.

    :param doc: Revit Document
    :type doc: Autodesk.Revit.DB.Document
    :param source_dims_by_view_name: Dictionary containing the source annotations by target view name. ( not all annotations from the active view may want to be copied to all views )
    :type source_dims_by_view_name: dict
    :param active_view: The active view containing all source annotations
    :type active_view: Autodesk.Revit.DB.View
    :param views_by_name_selected: Dictionary containing the selected views by name
    :type views_by_name_selected: dict
    :param callback_progress: Progress callback
    :type callback_progress: ProgressBase
    :return: Result object
    :rtype: Result
    """

    if isinstance(doc, Document) == False:
        raise ValueError("doc is not a Document")
    if isinstance(active_view, View) == False:
        raise ValueError("active_view is not a View")
    if isinstance(source_dims_by_view_name, dict) == False:
        raise ValueError("source_dims_by_view_name is not a dict")

    return_value = res.Result()

    t = Timer()
    t.start()

    # progress bar data
    max = len(source_dims_by_view_name)
    counter = 1

    # loop over target views and copy specific annotation form active view to it
    for target_view_name, source_dims in source_dims_by_view_name.items():

        # update the progress bar
        counter += 1
        if callback_progress:
            callback_progress.update(counter, max)

        # only copy unique dims ( when a string of dimensions is associated with a number of elements it may occur more then once in the list of dims to copy)
        dim_ids = []
        dim_ids_int = []
        for dim in source_dims:
            if dim.Id.IntegerValue not in dim_ids_int:
                # add dim to list
                dim_ids.append(dim.Id)
                dim_ids_int.append(dim.Id.IntegerValue)

        # cast to .net list of Autodesk.Revit.DB.ElementId
        dim_ids = dim_ids.ToList[ElementId]()

        # get the target view element from the dictionary by its name
        target_view = views_by_name_selected[target_view_name]

        def action():
            action_return_value = res.Result()
            try:
                # Define Transform and CopyPasteOptions()
                transform = Transform.Identity
                opts = CopyPasteOptions()
                # copy elements to other view
                ElementTransformUtils.CopyElements(
                    active_view,
                    dim_ids,
                    target_view,
                    transform,
                    opts,
                )
                action_return_value.append_message(
                    "Copied {} annotations(s) from view: {} to view: {}".format(
                        len(dim_ids),
                        active_view.Name,
                        target_view.Name,
                    )
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to copy {} annotations(s) from view: {} to view: {} with exception: {}".format(
                        len(dim_ids),
                        active_view.Name,
                        target_view.Name,
                        e,
                    ),
                )
            return action_return_value

        transaction = Transaction(
            doc,
            "Copy annotations {} to view: {}".format(len(dim_ids), target_view_name),
        )

        # TODO: replace with transaction suppression version
        transaction_result = in_transaction(transaction, action)
        # return views where annotations where added successfully so existing annotations can be deleted later on if required
        if transaction_result.status == True:
            return_value.result.append(target_view_name)

        # update the return value
        return_value.update(transaction_result)

        # check if cancelled
        if callback_progress and callback_progress.is_cancelled() == True:
            return_value.append_message("Cancelled by user")
            break

    return_value.append_message("{}".format(t.stop()))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions for Room separation lines.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from Autodesk.Revit.DB import (
    BuiltInCategory,
    ElementId,
    FilteredElementCollector,
    ModelCurve,
)

from duHast.Revit.Common.phases import get_all_phases_in_order
from duHast.Revit.Warnings.warning_guids import (
    ROOM_LINE_OFF_AXIS,
    ROOM_AND_WALL_SEPARATION_LINE_OVERLAP,
    ROOM_SEPARATION_LINES_OVERLAP,
)
from duHast.Revit.Warnings.warnings import get_unique_warnings_elements_by_guid


def get_room_separation_lines(doc):
    """
    Get all room separation lines in the document.

    :param doc: The Revit document.
    :type doc: Document

    :return: A collector containing room separation lines.
    :rtype: FilteredElementCollector

    """

    return FilteredElementCollector(doc).OfCategory(
        BuiltInCategory.OST_RoomSeparationLines
    )


def _remove_wall_ids(doc, ids):
    """
    Filters list of ids and returns ids of room separation lines (ModelCurve) only

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: List of element ids to filter
    :type ids: [Autodesk.Revit.DB.ElementId]

    :return: List of element ids
    :rtype: [Autodesk.Revit.DB.ElementId]

    """

    filtered_ids = []
    for id in ids:
        element = doc.GetElement(id)
        if isinstance(element, ModelCurve):
            filtered_ids.append(id)
    return filtered_ids


def get_all_room_separation_lines_ids_with_warnings(doc):
    """
    Returns the room separation line ids which have warnings associated with them.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids
    :rtype: [Autodesk.Revit.DB.ElementId]

    """

    # get room lines with warnings
    line_ids_of_axis = get_unique_warnings_elements_by_guid(
        doc=doc, guid=ROOM_LINE_OFF_AXIS
    )
    line_ids_overlapping = get_unique_warnings_elements_by_guid(
        doc=doc, guid=ROOM_SEPARATION_LINES_OVERLAP
    )

    # this will include wall ids...
    line_ids_overlapping_walls = get_unique_warnings_elements_by_guid(
        doc=doc, guid=ROOM_AND_WALL_SEPARATION_LINE_OVERLAP
    )
    # remove wall ids
    filtered_line_ides_overlapping_walls = _remove_wall_ids(
        doc=doc, ids=line_ids_overlapping_walls
    )

    # combine all ids
    all_line_ids_with_warnings = []
    if len(line_ids_of_axis) > 0:
        all_line_ids_with_warnings = all_line_ids_with_warnings + line_ids_of_axis
    if len(line_ids_overlapping) > 0:
        all_line_ids_with_warnings = all_line_ids_with_warnings + line_ids_overlapping
    if len(filtered_line_ides_overlapping_walls) > 0:
        all_line_ids_with_warnings = (
            all_line_ids_with_warnings + filtered_line_ides_overlapping_walls
        )

    return all_line_ids_with_warnings


def get_all_room_separation_lines_ids_without_warnings(doc):
    """
    returns the ids of all room separation lines in the model without any wanring associated with them.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids
    :rtype: [Autodesk.Revit.DB.ElementId]

    """

    all_room_separation_lines = get_room_separation_lines(doc=doc)
    all_line_ids_with_warnings = get_all_room_separation_lines_ids_with_warnings(
        doc=doc
    )

    all_room_separation_line_ids_without_warnings = []

    for room_sep in all_room_separation_lines:
        if room_sep.Id not in all_line_ids_with_warnings:
            all_room_separation_line_ids_without_warnings.append(room_sep.Id)

    return all_room_separation_line_ids_without_warnings


def get_room_separation_lines_by_level_name(doc, level_name):
    """
    Get all room separation lines in the document on a specific level by name.

    :param doc: The Revit document.
    :type doc: Document
    :param level_name: The name of the level.
    :type level_name: str

    :return: A list of room separation lines on the specified level.
    :rtype: list

    """

    room_separation_lines = []
    # get all room separation lines in the model
    col = get_room_separation_lines(doc)

    # loop over lines and filter them by level name
    for room_line in col:
        level_name_from_line = doc.GetElement(room_line.LevelId).Name
        if level_name_from_line == level_name:
            room_separation_lines.append(room_line)

    # return any matches
    return room_separation_lines


def get_room_separation_lines_by_level_names(doc):
    """
    Get all room separation lines in the document sorted by level name.

    :param doc: The Revit document.
    :type doc: Document

    :return: A dictionary containing room separation lines sorted by level name.
    :rtype: dict

    """

    room_separation_lines = {}

    # get all room separation lines in the model
    col = get_room_separation_lines(doc)

    # loop over lines and sort them by level name
    for room_line in col:
        level_name_from_line = doc.GetElement(room_line.LevelId).Name
        if level_name_from_line in room_separation_lines:
            room_separation_lines[level_name_from_line].append(room_line)
        else:
            room_separation_lines[level_name_from_line] = [room_line]

    # return any sorted lines
    return room_separation_lines


def sort_room_separation_line_by_phase_created(room_separation_lines):
    """
    Sorts the room separation lines by phase created id

    :param room_separation_lines: list of room separation lines
    :type room_separation_lines:list

    :return: A dictionary of room separation lines, where key is the phase created id, and values are the associated room separation lines
    :rtype: {ElementId:[ModelLine]}

    """

    sorted_lines = {}

    for line in room_separation_lines:
        if line.CreatedPhaseId in sorted_lines:
            sorted_lines[line.CreatedPhaseId].append(line)
        else:
            sorted_lines[line.CreatedPhaseId] = [line]

    return sorted_lines


def filter_room_separation_lines_by_phase_created(
    doc, room_separation_lines, phase_created_name
):
    """
    Filter room separation lines by phase created.

    :param doc: The Revit document.
    :type doc: Document
    :param phase_created_name: The phase name of the phase the room separation lines was created in.
    :type phase_created_name: str

    :return: A list of room separation lines filtered by phase created.
    :rtype: list

    """

    room_separation_lines_filtered = []

    # loop over lines and filter them by phase created
    for room_line in room_separation_lines:
        phase_Line_created_name = doc.GetElement(room_line.CreatedPhaseId).Name
        # print(" {} vs {} ".format(phase_Line_created_name, phase_created_name))
        if phase_Line_created_name == phase_created_name:
            room_separation_lines_filtered.append(room_line)

    # return any matches
    return room_separation_lines_filtered


def filter_room_separation_lines_by_phase_demolished(
    doc, room_separation_lines, phase_demolished_name
):
    """
    Filter room separation lines by phase demolished.

    :param doc: The Revit document.
    :type doc: Document
    :param phase_demolished_name: The phase name of the phase the room separation lines was demolished in.
    :type phase_demolished_name: str

    :return: A list of room separation lines filtered by phase demoplished.
    :rtype: list

    """

    room_separation_lines_filtered = []

    # loop over lines and filter them by phase created
    for room_line in room_separation_lines:

        # get the demo phase of the room separation line
        phase_Line_demoed = doc.GetElement(room_line.DemolishedPhaseId)

        # check if both values are None (no demoed phase set)
        if phase_Line_demoed is None and phase_demolished_name == None:
            room_separation_lines_filtered.append(room_line)
        # check if one value is None (no demoed phase set but checking for demoed phase name)
        elif phase_Line_demoed is None and phase_demolished_name != None:
            continue
        # check if one value is None (a demoed phase is set but checking for no demoed phase)
        elif phase_Line_demoed is not None and phase_demolished_name == None:
            continue
        # check if both values are not None ( a demoed phase is set and checking for demoed phase name )
        else:
            phase_line_demolished_name = phase_Line_demoed.Name
            if phase_line_demolished_name == phase_demolished_name:
                room_separation_lines_filtered.append(room_line)

    # return any matches
    return room_separation_lines_filtered


def filter_room_separation_lines_by_phase_created_older_than_upper_bound_phase(
    doc, room_separation_lines, upper_bound_phase_id
):
    """
    Filter room separation lines by phase created. Phase created value must be a phase equal to or older than the upper bound phase id.
    If upper bound phase Id is and ElementId.InvalidElementId the list will be returned unchanged.

    :param doc: The Revit document.
    :type doc: Document
    :param room_separation_lines: A list of room separation lines
    :type room_separation_lines: []
    :param upper_bound_phase_id: the id of the phase to be used as upper bound
    :type upper_bound_phase_id: Autodesk.Revit.ElementId

    :return: A list of room separation lines filtered by phase created.
    :rtype: list

    """

    if isinstance(upper_bound_phase_id, ElementId) == False:
        raise TypeError(
            "upper_bound_phase_id needs to be of type ElementId. But is: {}".format(
                upper_bound_phase_id
            )
        )

    if upper_bound_phase_id == ElementId.InvalidElementId:
        return room_separation_lines

    # filtered lines
    lines_filtered = []

    # phases in model
    phases_ordered = get_all_phases_in_order(doc)

    # flip the phases_ordered list to get the index of the active view phase
    phase_id_index_by_phase_id = {}
    counter = 0
    for phase in phases_ordered:
        phase_id_index_by_phase_id[phase[0]] = counter
        counter += 1

    # index of upper bound
    upper_bound_phase_id_index = phase_id_index_by_phase_id[upper_bound_phase_id]

    # loop over lines and filter away any room separation lines newer then the upper bound phase id
    for line in room_separation_lines:
        line_phase_index = phase_id_index_by_phase_id[line.CreatedPhaseId]
        if line_phase_index <= upper_bound_phase_id_index:
            lines_filtered.append(line)

    return lines_filtered


def filter_room_separation_lines_by_level_id(room_separation_lines, level_id):
    """
    Filter room separation lines by level id its placed on.

    :param room_separation_lines: A list of room separation lines
    :type room_separation_lines: []
    :param level_id: The id of he level to be filtered by.
    :type level_id: Autodesk.Revit.ElementId

    :return: A list of room separation lines filtered by level placed on.
    :rtype: list

    """

    lines_filtered = []
    for room_line in room_separation_lines:
        if room_line.LevelId == level_id:
            lines_filtered.append(room_line)

    return lines_filtered



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.LibraryCleanUp.Utility.defaults import SWAP_DIRECTIVE_FILE_NAME, MAINTAIN_TYPES_BY_FAMILY_FILE_NAME_PREFIX
from duHast.Utilities.files_get import get_files_single_directory

from duHast.Utilities.files_csv import read_csv_file

def read_maintain_types (directory_path):
    return_value = Result()

    try:

        # get type maintain file path
        files = get_files_single_directory(directory_path,MAINTAIN_TYPES_BY_FAMILY_FILE_NAME_PREFIX, "", ".csv")
           
        if len(files) == 0:
            return_value.update_sep(
                False,
                "No maintain directives found in directory: {}".format(directory_path),
            )
            return return_value
        
        # build the file path for the maintain directives
        file_path = files[0]  # assuming we take the first file found

        return_value.append_message("Reading maintain directives to file: {}".format(file_path))

        # write the maintain directives to the file
        csv_file_result = read_csv_file(file_path=file_path)

        return_value.update(csv_file_result)
        
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to read maintain directives with exception: {}".format(e),
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of lists containing loadable family categories. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import BuiltInCategory
from System.Collections.Generic import List

#: This list contains 3D element categories and is used in obsolete revit family purge function
#: any revit category commented out with note 'purged else where' can be found in list 'catsLoadableThreeDOther'
CATEGORIES_LOADABLE_3D = List[BuiltInCategory](
    [
        # BuiltInCategory.OST_CableTrayFitting,  purged else where
        BuiltInCategory.OST_Casework,
        BuiltInCategory.OST_Columns,
        BuiltInCategory.OST_CommunicationDevices,
        # BuiltInCategory.OST_ConduitFitting,  purged else where
        # BuiltInCategory.OST_CurtainWallPanels, purged else where
        BuiltInCategory.OST_DataDevices,
        # BuiltInCategory.OST_DetailComponents, purged else where
        BuiltInCategory.OST_Doors,
        # BuiltInCategory.OST_DuctAccessory,  purged else where
        # BuiltInCategory.OST_DuctTerminal, purged else where
        # BuiltInCategory.OST_DuctFitting,  purged else where
        BuiltInCategory.OST_ElectricalEquipment,
        BuiltInCategory.OST_ElectricalFixtures,
        BuiltInCategory.OST_Entourage,
        BuiltInCategory.OST_FireAlarmDevices,
        BuiltInCategory.OST_Furniture,
        BuiltInCategory.OST_FurnitureSystems,
        BuiltInCategory.OST_GenericModel,
        BuiltInCategory.OST_LightingFixtures,
        BuiltInCategory.OST_LightingDevices,
        BuiltInCategory.OST_MechanicalEquipment,
        BuiltInCategory.OST_NurseCallDevices,
        BuiltInCategory.OST_Parking,
        # BuiltInCategory.OST_PipeAccessory,  purged else where
        # BuiltInCategory.OST_PipeFitting,  purged else where
        BuiltInCategory.OST_Planting,
        BuiltInCategory.OST_PlumbingFixtures,
        # BuiltInCategory.OST_ProfileFamilies, #purged elsewhere
        BuiltInCategory.OST_SecurityDevices,
        BuiltInCategory.OST_Site,
        BuiltInCategory.OST_SpecialityEquipment,
        BuiltInCategory.OST_Sprinklers,
        # BuiltInCategory.OST_StairsRailingBaluster, #purged else where
        BuiltInCategory.OST_StructuralColumns,
        BuiltInCategory.OST_StructuralFoundation,
        BuiltInCategory.OST_StructuralFraming,
        BuiltInCategory.OST_TitleBlocks,
        BuiltInCategory.OST_TelephoneDevices,
        BuiltInCategory.OST_Windows,
    ]
)

#: Contains 3D family categories which needed specific purge code, rather then checking for unplaced family instances.
#: i.e. built in revit type settings
CATEGORIES_LOADABLE_3D_OTHER = List[BuiltInCategory](
    [
        BuiltInCategory.OST_CableTrayFitting,
        BuiltInCategory.OST_ConduitFitting,
        BuiltInCategory.OST_CurtainWallPanels,
        BuiltInCategory.OST_DetailComponents,
        BuiltInCategory.OST_DuctAccessory,
        BuiltInCategory.OST_DuctTerminal,
        BuiltInCategory.OST_DuctFitting,
        BuiltInCategory.OST_PipeAccessory,
        BuiltInCategory.OST_PipeFitting,
        BuiltInCategory.OST_ProfileFamilies,
        BuiltInCategory.OST_StairsRailingBaluster,
    ]
)

# categories which got added in Revit 2022
CATEGORIES_LOADABLE_3D_REVIT_2022 = List[BuiltInCategory]([
    BuiltInCategory.OST_FoodServiceEquipment,
    BuiltInCategory.OST_MedicalEquipment,
    BuiltInCategory.OST_FireProtection,
    BuiltInCategory.OST_VerticalCirculation,
    BuiltInCategory.OST_AudioVisualDevices,
    BuiltInCategory.OST_Signage,
    BuiltInCategory.OST_Hardscape,
    BuiltInCategory.OST_TemporaryStructure,
    BuiltInCategory.OST_Roads,
    BuiltInCategory.OST_BridgeAbutments,
    BuiltInCategory.OST_BridgeBearings,
    BuiltInCategory.OST_BridgePiers,
    BuiltInCategory.OST_BridgeCables,
    BuiltInCategory.OST_BridgeDecks,
    BuiltInCategory.OST_ExpansionJoints,
    BuiltInCategory.OST_StairsRailing,
    BuiltInCategory.OST_RailingTermination,
    BuiltInCategory.OST_RailingSupport,
])


#: This list contains 2D element categories and is used in obsolete revit family purge function.\
#: any revit category commented out with note 'purged else where' can be found in list 'catsLoadableTagsOther'
CATEGORIES_LOADABLE_TAGS = List[BuiltInCategory](
    [
        BuiltInCategory.OST_CurtainWallPanelTags,
        BuiltInCategory.OST_AreaTags,
        BuiltInCategory.OST_CaseworkTags,
        # BuiltInCategory.OST_CalloutHeads, #purged separately
        BuiltInCategory.OST_CeilingTags,
        BuiltInCategory.OST_DataDeviceTags,
        BuiltInCategory.OST_DetailComponentTags,
        BuiltInCategory.OST_DoorTags,
        BuiltInCategory.OST_DuctAccessoryTags,
        BuiltInCategory.OST_DuctFittingTags,
        BuiltInCategory.OST_DuctInsulationsTags,
        BuiltInCategory.OST_DuctLiningsTags,
        BuiltInCategory.OST_DuctTags,
        BuiltInCategory.OST_DuctTerminalTags,
        BuiltInCategory.OST_ElectricalCircuitTags,
        BuiltInCategory.OST_ElectricalEquipmentTags,
        BuiltInCategory.OST_ElectricalFixtureTags,
        # BuiltInCategory.OST_ElevationMarks, #purged separately
        BuiltInCategory.OST_FabricAreaTags,
        BuiltInCategory.OST_FabricReinforcementTags,
        BuiltInCategory.OST_FireAlarmDeviceTags,
        BuiltInCategory.OST_FlexDuctTags,
        BuiltInCategory.OST_FlexPipeTags,
        BuiltInCategory.OST_FloorTags,
        BuiltInCategory.OST_FoundationSlabAnalyticalTags,
        BuiltInCategory.OST_FurnitureTags,
        BuiltInCategory.OST_FurnitureSystemTags,
        BuiltInCategory.OST_GenericModelTags,
        # BuiltInCategory.OST_GenericAnnotation, # purged separately tricky one...some of these might be used in dimensions for instance...
        # BuiltInCategory.OST_GridHeads, # purged separately
        BuiltInCategory.OST_InternalAreaLoadTags,
        BuiltInCategory.OST_InternalLineLoadTags,
        BuiltInCategory.OST_InternalPointLoadTags,
        BuiltInCategory.OST_IsolatedFoundationAnalyticalTags,
        BuiltInCategory.OST_KeynoteTags,
        # BuiltInCategory.OST_LevelHeads, #purged separately
        BuiltInCategory.OST_LightingDeviceTags,
        BuiltInCategory.OST_LightingFixtureTags,
        BuiltInCategory.OST_LineLoadTags,
        BuiltInCategory.OST_LinkAnalyticalTags,
        BuiltInCategory.OST_MassTags,
        BuiltInCategory.OST_MaterialTags,
        BuiltInCategory.OST_MechanicalEquipmentTags,
        BuiltInCategory.OST_MEPSpaceTags,
        BuiltInCategory.OST_MultiCategoryTags,
        BuiltInCategory.OST_NodeAnalyticalTags,
        BuiltInCategory.OST_NurseCallDeviceTags,
        BuiltInCategory.OST_ParkingTags,
        BuiltInCategory.OST_PartTags,
        BuiltInCategory.OST_PathReinTags,
        BuiltInCategory.OST_PipeAccessoryTags,
        BuiltInCategory.OST_PipeFittingTags,
        BuiltInCategory.OST_PipeInsulationsTags,
        BuiltInCategory.OST_PipeTags,
        BuiltInCategory.OST_PlantingTags,
        BuiltInCategory.OST_PlumbingFixtureTags,
        BuiltInCategory.OST_RailingSystemTags,
        BuiltInCategory.OST_RebarTags,
        # BuiltInCategory.OST_ReferenceViewerSymbol, #purged separately
        BuiltInCategory.OST_RevisionCloudTags,
        BuiltInCategory.OST_RoofTags,
        BuiltInCategory.OST_RoomTags,
        # BuiltInCategory.OST_SectionHeads, #purged separately
        BuiltInCategory.OST_SecurityDeviceTags,
        BuiltInCategory.OST_SitePropertyLineSegmentTags,
        BuiltInCategory.OST_SitePropertyTags,
        BuiltInCategory.OST_SpecialityEquipmentTags,
        # BuiltInCategory.OST_SpotElevSymbols, #purged elsewhere
        BuiltInCategory.OST_SprinklerTags,
        BuiltInCategory.OST_StairsLandingTags,
        BuiltInCategory.OST_StairsRailingTags,
        BuiltInCategory.OST_StairsRunTags,
        BuiltInCategory.OST_StairsSupportTags,
        BuiltInCategory.OST_StairsTags,
        BuiltInCategory.OST_StairsTriserTags,
        BuiltInCategory.OST_StructConnectionTags,
        BuiltInCategory.OST_StructuralColumnTags,
        BuiltInCategory.OST_StructuralFoundationTags,
        BuiltInCategory.OST_StructuralFramingTags,
        BuiltInCategory.OST_StructuralStiffenerTags,
        BuiltInCategory.OST_TelephoneDeviceTags,
        BuiltInCategory.OST_TrussTags,
        # BuiltInCategory.OST_ViewportLabel, #purged elsewhere
        BuiltInCategory.OST_WallTags,
        BuiltInCategory.OST_WindowTags,
    ]
)

#: Contains 2D family categories which needed specific purge code, rather then checking for unplaced family instances
#: i.e. built in revit type settings
CATEGORIES_LOADABLE_TAGS_OTHER = List[BuiltInCategory](
    [
        BuiltInCategory.OST_CalloutHeads,
        BuiltInCategory.OST_ElevationMarks,
        BuiltInCategory.OST_GenericAnnotation,
        BuiltInCategory.OST_GridHeads,
        BuiltInCategory.OST_LevelHeads,
        BuiltInCategory.OST_ReferenceViewerSymbol,
        BuiltInCategory.OST_SectionHeads,
        BuiltInCategory.OST_SpotElevSymbols,
        BuiltInCategory.OST_ViewportLabel,
    ]
)

# categories which got added in Revit 2022
CATEGORIES_LOADABLE_TAGS_REVIT_2022 = List[BuiltInCategory]([
    BuiltInCategory.OST_FoodServiceEquipmentTags,
    BuiltInCategory.OST_MedicalEquipmentTags,
    BuiltInCategory.OST_FireProtectionTags,
    BuiltInCategory.OST_VerticalCirculationTags,
    BuiltInCategory.OST_AudioVisualDeviceTags,
    BuiltInCategory.OST_SignageTags,
    BuiltInCategory.OST_HardscapeTags,
    BuiltInCategory.OST_TemporaryStructureTags,
    BuiltInCategory.OST_RoadTags,
    BuiltInCategory.OST_BridgeBearingTags,
    BuiltInCategory.OST_BridgePierTags,
    BuiltInCategory.OST_BridgeCableTags,
    BuiltInCategory.OST_BridgeDeckTags,
    BuiltInCategory.OST_AbutmentWallTags,	
    BuiltInCategory.OST_AbutmentPileTags,
    BuiltInCategory.OST_AbutmentFoundationTags,
    BuiltInCategory.OST_ExpansionJointTags,
])



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Polygon geometry data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A polygon consists, as a minimum, of an outer loop, but may also have any number of inner loops. Those inner loops describe holes in the surface the outer loop decribes.

Loops are made up of a number of 2D points.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Data.Objects.Collectors.Properties.Geometry import geometry_base
from duHast.Data.Objects.Collectors.Properties.data_property_names import DataPropertyNames
from duHast.Geometry.point_2 import Point2


class DataGeometryPolygon2(geometry_base.DataGeometryBase):
    data_type = "polygon"

    def __init__(self, j=None):
        """
        Class constructor for a 2d polygon.

        :param j:  json formatted dictionary of this class, defaults to {}
        :type j: dict, optional
        """

        # store data type  in base class and pass on json string!!!
        super(DataGeometryPolygon2, self).__init__(DataGeometryPolygon2.data_type, j)

        # set default values
        self.outer_loop = []
        self.inner_loops = []

        json_var = None
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # no action required
                json_var = j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )

            # attempt to populate from json
            try:
                # get outer points loop
                outer_loop = json_var.get(DataPropertyNames.OUTER_LOOP, [])

                # need a minimum of 3 points to form a polygon
                if len(outer_loop) >= 3:
                    for p in outer_loop:
                        self.outer_loop.append(Point2(j=p))
                elif 3 > len(outer_loop) > 0:
                    # not enough points
                    raise ValueError(
                        "outer loop data needs to contain at least 3 points"
                    )
                else:
                    # a polygon must contain at least an outer loop...
                    raise ValueError("Json did not contain any outer loop data")

                # get inner loops
                inner_loops = json_var.get(DataPropertyNames.INNER_LOOPS, [])
                if len(inner_loops) > 0:
                    for loop in inner_loops:
                        loop_points = []
                        for p in loop:
                            loop_points.append(Point2(j=p))
                        self.inner_loops.append(loop_points)

            except Exception as e:
                msg = "Node {} failed to initialise with: {}".format(self.data_type, e)
                raise type(e)(msg)

    def add_point_to_outer_loop(self, point):
        """
        Adds a point to the outer loop.

        :param point: Point2 instance to add to the outer loop
        :type point: Point2
        :raises TypeError: if the point is not an instance of Point2
        """

        if not isinstance(point, Point2):
            raise TypeError("Point must be an instance of Point2.")
        self.outer_loop.append(point)

    def add_inner_loop(self, loop):
        """
        Adds a new inner loop to the inner loops list.

        :param loop: List of Point2 instances representing an inner loop
        :type loop: list[Point2]
        :raises ValueError: if the loop does not contain at least 3 points
        :raises TypeError: if any element in the loop is not an instance of Point2
        """

        if len(loop) < 3:
            raise ValueError("An inner loop must contain at least 3 points.")
        if not all(isinstance(point, Point2) for point in loop):
            raise TypeError("All points in the loop must be instances of Point2.")
        self.inner_loops.append(loop)

    def __eq__(self, other):
        if not isinstance(other, DataGeometryPolygon2):
            return NotImplemented
        return (
            self.outer_loop == other.outer_loop
            and self.inner_loops == other.inner_loops
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.outer_loop, self.inner_loops))


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from duHast.Utilities.Objects.result import Result
from duHast.Utilities.unit_conversion import convert_mm_to_imperial_feet
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import BuiltInParameter, Ceiling,ElementId, Transaction



def create_ceiling( doc, level_id, outline, elevation, ceiling_type_id, phase_created = None, transaction_manager=in_transaction):
    """
    Creates a ceiling in the given document at the specified level and elevation.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param level_id: The ElementId of the level where the ceiling should be created.
    :type level_id: Autodesk.Revit.DB.ElementId
    :param outline: The outline of the ceiling as a list of curve loops.
    :type outline: Autodesk.Revit.DB.CurveLoop
    :param elevation: The elevation in mmm of the ceiling above the level.
    :type elevation: float
    :param ceiling_type_id: The ElementId of the ceiling type to use.
    :type ceiling_type_id: Autodesk.Revit.DB.ElementId
    :param transaction_manager: Optional transaction manager to handle transactions.
    :type transaction_manager: function
    
    :return:
        Result class instance.

        - Ceiling creation status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property contains id of ceiling created.
        - Result.result will contain the created ceiling element if successful, otherwise an empty list.

        On exception:

        - .status (bool) will be False.
        - .message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = Result()
  
    def action ():
        action_return_value = Result()
        try:
            ceiling = Ceiling.Create(doc, outline, ElementId.InvalidElementId, level_id)

            if not ceiling:
                action_return_value.update_sep(False, "failed to create ceiling")
                return action_return_value
            
            # set the ceiling type
            ceiling.ChangeTypeId(ceiling_type_id)

            # set the elevation ( expectation is that the elevation is given in mm )
            param = ceiling.get_Parameter(BuiltInParameter.CEILING_HEIGHTABOVELEVEL_PARAM)
            param.Set(convert_mm_to_imperial_feet(elevation))

            # set the phase created if there is one provided
            if phase_created:
                # set the phase created if there is one provided
                ceiling.get_Parameter(BuiltInParameter.PHASE_CREATED).Set(phase_created.Id)
           
            # add the ceiling element to the return object
            action_return_value.result.append(ceiling)
            action_return_value.append_message("created ceiling: {}".format(ceiling.Id))
            return action_return_value
        except Exception as e:
            action_return_value.update_sep(False, "failed to create ceiling: {}".format(e))
            return action_return_value

    # check if a transaction manager is provided
    if transaction_manager:
        # run the action in a transaction
        tranny = Transaction(doc, "Adding ceiling")
        return_value = transaction_manager(tranny, action)
    else:
        return_value.update(action())
       

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Class for family base data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json

from duHast.Revit.Family.Data.Objects import ifamily_data_storage as IFamDataStorage
from duHast.Revit.Categories.Data.Objects.category_data_storage_used_by import (
    FamilyCategoryDataStorageUsedBy,
)


class FamilyCategoryDataStorage(IFamDataStorage.IFamilyDataStorage):

    # data type for this class ( used in reports as first entry per row )
    data_type = "Category"

    # number of properties in this class ( used in report reader function )
    number_of_properties = 20

    def __init__(
        self,
        root_name_path,
        root_category_path,
        family_name,
        family_file_path,
        use_counter,
        used_by,
        category_name,
        sub_category_name,
        sub_category_id,
        category_graphics_style_three_d,
        category_graphics_style_cut,
        category_graphics_style_projection,
        property_material_name,
        property_material_id,
        property_line_weight_cut_name,
        property_line_weight_projection_name,
        property_line_colour_red_name,
        property_line_colour_green_name,
        property_line_colour_blue,
        **kwargs
    ):

        # store args in base class
        super(FamilyCategoryDataStorage, self).__init__(
            data_type=FamilyCategoryDataStorage.data_type,
            root_name_path=root_name_path,
            root_category_path=root_category_path,
            family_name=family_name,
            family_file_path=family_file_path,
        )

        # check whether used by is a json list as a string
        if used_by is not None and used_by != "None":
            # check type of data that came in:
            # check if string was passed in
            if isinstance(used_by, str):
                # a string
                used_by_json = json.loads(used_by)
                # if used by was converted into a json object successfully it should now be a list of dictionaries!
                converted_used_by = []
                if isinstance(used_by_json, list):
                    for i in range(len(used_by_json)):
                        try:
                            # convert to FamilyCategoryDataStorageUsedBy object
                            dummy = FamilyCategoryDataStorageUsedBy(j=used_by_json[i])
                            converted_used_by.append(dummy)
                        except Exception as e:
                            raise ValueError(
                                "Failed to convert string {} to FamilyCategoryDataStorageUsedBy object: {}".format(
                                    used_by_json[i], e
                                )
                            )
                    # override string with converted data
                    used_by = converted_used_by
            # check if a list of family category data storage used by objects was passed in
            elif isinstance(used_by, list):
                # check if every list item is of type FamilyCategoryDataStorageUsedBy
                for i in range(len(used_by)):
                    if not isinstance(used_by[i], FamilyCategoryDataStorageUsedBy):
                        raise ValueError(
                            "used_by list must contain only FamilyCategoryDataStorageUsedBy objects"
                        )
            elif isinstance(used_by, FamilyCategoryDataStorageUsedBy):
                # convert to a list
                used_by = [used_by]
            else:
                raise ValueError(
                    "used_by argument supplied must be of type string or type FamilyCategoryDataStorageUsedBy or a list of FamilyCategoryDataStorageUsedBy"
                )
        else:
            # default is an empty list
            used_by = []

        # store other args in this class
        self.use_counter = use_counter
        self.used_by = used_by
        self.category_name = category_name
        self.sub_category_name = sub_category_name
        self.sub_category_id = sub_category_id
        self.category_graphics_style_3d = category_graphics_style_three_d
        self.category_graphics_style_cut = category_graphics_style_cut
        self.category_graphics_style_projection = category_graphics_style_projection
        self.property_material_name = property_material_name
        self.property_material_id = property_material_id
        self.property_line_weight_cut_name = property_line_weight_cut_name
        self.property_line_weight_projection_name = property_line_weight_projection_name
        self.property_line_colour_red_name = property_line_colour_red_name
        self.property_line_colour_green_name = property_line_colour_green_name
        self.property_line_colour_blue_name = property_line_colour_blue

    def __eq__(self, other):
        """
        Custom compare is equal override.

        :param other: Another instance of FamilyCategoryDataStorage base class
        :type other: :class:`.FamilyCategoryDataStorage`
        :return: True if all properties of compared class instances are equal, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, FamilyCategoryDataStorage) and (
            self.data_type,
            self.root_name_path,
            self.root_category_path,
            self.family_name,
            self.family_file_path,
            self.use_counter,
            self.used_by,
            self.category_name,
            self.sub_category_name,
            self.sub_category_id,
            self.category_graphics_style_3d,
            self.category_graphics_style_cut,
            self.category_graphics_style_projection,
            self.property_material_name,
            self.property_material_id,
            self.property_line_weight_cut_name,
            self.property_line_weight_projection_name,
            self.property_line_colour_red_name,
            self.property_line_colour_green_name,
            self.property_line_colour_blue_name,
        ) == (
            other.data_type,
            other.root_name_path,
            other.root_category_path,
            other.family_name,
            other.family_file_path,
            other.use_counter,
            other.used_by,
            other.category_name,
            other.sub_category_name,
            other.sub_category_id,
            other.category_graphics_style_3d,
            other.category_graphics_style_cut,
            other.category_graphics_style_projection,
            other.property_material_name,
            other.property_material_id,
            other.property_line_weight_cut_name,
            other.property_line_weight_projection_name,
            other.property_line_colour_red_name,
            other.property_line_colour_green_name,
            other.property_line_colour_blue_name,
        )

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to fields in Revit view schedules. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import ElementId, SectionType

# the sheet number parameter id
SHEET_NUMBER_PARAMETER_ID = ElementId(-1007401)

def schedule_contains_field_by_parameter_id(schedule, parameter_id, ignore_hidden_field=False):
    """
    Checks if the given schedule contains a specific field identified by the parameter Id.

    :param schedule: The schedule to check.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    :param parameter_id: The parameter Id of the field to check for.
    :type parameter_id: Autodesk.Revit.DB.ElementId
    :param ignore_hidden_field: If True, a hidden field returns false.
    :type ignore_hidden_field: bool, optional
    
    :return: True if the schedule contains the sheet number field, False otherwise.
    :rtype: bool
    """

    # get the schedule definition
    schedule_definition = schedule.Definition

    # get the number of fields in the schedule
    num_fields = schedule_definition.GetFieldCount()

    # loop through the fields and get their names
    for i in range(num_fields):
        field = schedule_definition.GetField(i)
        field_id = field.ParameterId
        # check if the field id matches the parameter id
        if field_id == parameter_id:
            if ignore_hidden_field is True and field.IsHidden:
                # if the field is hidden, return false
                return False
            # return True 
            return True
       
    # field not found, return False
    return False


def schedule_contains_sheet_number_field(schedule, ignore_hidden_field=False):
    """
    Checks if the given schedule contains the sheet number field.

    :param schedule: The schedule to check.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    :param ignore_hidden_field: If True, a hidden sheet number field returns false.
    :type ignore_hidden_field: bool, optional

    :return: True if the schedule contains the sheet number field, False otherwise.
    :rtype: bool
    """

    return schedule_contains_field_by_parameter_id(schedule, SHEET_NUMBER_PARAMETER_ID, ignore_hidden_field=ignore_hidden_field)
   

def get_field_names_from_schedule(schedule):
    """
    Get a list of all field names from a Revit schedule.
    :param schedule: The Revit schedule object from which to extract field names.
    :type schedule: Autodesk.Revit.DB.ViewSchedule

    :return: A list of field names in the schedule.
    :rtype: list of str
    """

    # get the schedule definition
    schedule_definition = schedule.Definition

    # get the number of fields in the schedule
    num_fields = schedule_definition.GetFieldCount()

    # create a list to hold the field names
    field_names = []

    # loop through the fields and get their names
    for i in range(num_fields):
        field = schedule_definition.GetField(i)
        field_name = field.GetName()
        field_names.append(field_name)
    
    return field_names


def get_field_names_to_parameters(schedule):
    """
    Get a dictionary of field names to their corresponding parameters in a Revit schedule.
    
    :param schedule: The Revit schedule object from which to extract field names and parameters.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    
    :return: A dictionary where keys are field names and values are the corresponding parameter IDs.
    """

    # create a dictionary to hold field names and their corresponding parameter IDs
    field_names_to_id = {}

    # get the schedule definition
    schedule_definition = schedule.Definition

    # get the number of fields in the schedule
    num_fields = schedule_definition.GetFieldCount()

    # loop through the fields and get their names
    for i in range(num_fields):
        field = schedule_definition.GetField(i)
        field_name = field.GetName()
        parameter_id = field.ParameterId

        # add the field name and parameter ID to the dictionary
        field_names_to_id[field_name]= parameter_id
    
    return field_names_to_id


def get_field_column_index_from_schedule_by_parameter_id(schedule, parameter_id):
    """
    Returns the index of a field in a Revit schedule based on the parameter ID.

    If the field is hidden or not found will return -1.

    Takes into account if other fields are hidden or not, so the index is based on visible fields only.

    :param schedule: The Revit schedule object from which to extract the field index.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    :param parameter_id: The parameter ID of the field to find.
    :type parameter_id: Autodesk.Revit.DB.ElementId

    :return: The index of the field in the schedule, or -1 if not found or hidden.
    :rtype: int 
    """


    # get the schedule definition
    schedule_definition = schedule.Definition

    # get the number of fields in the schedule
    num_fields = schedule_definition.GetFieldCount()
    
    # ordered field list:
    sorted_field_ids = schedule_definition.GetFieldOrder()

    # get the column index
    column_index = 0

    # if the parameter ID is not provided, return -1
    field_found = False

    # loop through the fields in order of appearance in schedule to find the column index for the specified parameter ID
    for i in range(num_fields):
        field_id = sorted_field_ids[i]

        # get the field
        field_by_field_id = schedule_definition.GetField(field_id)

        # check if the field is hidden
        if field_by_field_id.IsHidden:
            # if the field is hidden, skip it
            continue

        # check if the field parameter ID matches the specified parameter ID
        if field_by_field_id.ParameterId == parameter_id:
            field_found = True
            break

        # increase the column index if the field is not hidden
        column_index += 1
    
    # if the field was found, return the column index
    if not field_found:
        column_index = -1

    return column_index


def get_field_values_from_schedule_by_parameter_id(schedule, parameter_id=None):
    """
    Get a list of all field values from a Revit schedule based on the parameter id.

    
    :param schedule: The Revit schedule object from which to extract field values.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    
    :return: A list of field values in the schedule.
    :rtype: list of str
    """

    # get the schedule definition
    schedule_definition = schedule.Definition

    # get the number of fields in the schedule
    num_fields = schedule_definition.GetFieldCount()
    
    # ordered field list:
    sorted_field_ids = schedule_definition.GetFieldOrder()

    # get the column index
    column_index = -1

    # loop through the fields in order of appearance in schedule to find the column index for the specified parameter ID
    for i in range(num_fields):
        field_id = sorted_field_ids[i]
        # print("Field ID: {field_id}, Parameter ID: {parameter_id}".format(
        #     field_id=field_id,
        #     parameter_id=parameter_id
        # ))

        field_by_field_id = schedule_definition.GetField(field_id)
        #print("field by field ID",field_by_field_id.GetName())

        field_index = schedule_definition.GetField(i)
        #print("field by index",field_index.GetName())

        if field_by_field_id.ParameterId == parameter_id and field_by_field_id.IsHidden == False:
            column_index = i

            # print("Found column index: {column_index} for parameter ID: {parameter_id} for schedule {schedule_name}".format(
            #     column_index=column_index,
            #     parameter_id=parameter_id,
            #     schedule_name=schedule.Name
            # ))
            break

    if column_index == -1:
        raise ValueError("The specified parameter ID is not found in the schedule fields.")

    # get the table data
    table = schedule.GetTableData()

    # get the body table data
    table_body = table.GetSectionData(SectionType.Body)
  
    # get the number of rows in the body table 
    num_rows = table_body.NumberOfRows

    # create a list to hold the field values
    field_values = []

    for row in range(num_rows):
        # get the cell text for the specified row and column
        # get cell text expects the row number first, followed by the column number
        cell_text = table_body.GetCellText(row, column_index)
        print("Row: {row}, Column: {column_index}, Cell Text: {cell_text}".format(
            row=row,
            column_index=column_index,
            cell_text=cell_text
        ))
        # append the cell text to the field values list
        field_values.append(cell_text)
    
    print (field_values)
    return field_values


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_get import get_files_from_directory_walker_with_filters_simple
from duHast.Utilities.files_io import file_exist, file_delete

def remove_obsolete_part_atom_exports(process_dirs):
    """
    Removes all xml files in the process directories with no associated family files.

    :param process_dirs: List of directories to process.
    :type process_dirs: list
    :return: Result object with status and message.
    :rtype: Result
    """

    # set up a status tracker
    return_value = Result()

    try:

        # build list of xml files in process directories
        xml_files = []

        for d in process_dirs:
            return_value.append_message("Processing directories: {}".format(d))
        
            # get all xml files in the library location
            xml_in_directory = get_files_from_directory_walker_with_filters_simple(d, ".xml")
            if xml_in_directory and len(xml_in_directory) > 0:
                xml_files = xml_files + xml_in_directory
        
        xml_files_delete = []
        # check if any files are missing their matching family file
        for xml in xml_files:
            rfa_file = os.path.splitext(xml)[0] + ".rfa"
            if not file_exist(rfa_file):
                xml_files_delete.append(xml)
               
        # delete the xml files
        for xml in xml_files_delete:
            deleted_flag = file_delete(xml)
            if not deleted_flag:
                return_value.update_sep(
                    False, "Failed to delete xml file: {}".format(xml)
                )
            else:
                return_value.append_message(
                    "Deleted: {}".format(xml)
                )
    
    except Exception as e:
        return_value.update_sep(
            False, "Failed to delete files with exception: {}".format(e)
        )

    return_value.append_message("Finished")
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit walls properties report function. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Materials.materials import get_material_name_by_id
from duHast.Revit.Common.common import get_element_mark
from duHast.Revit.Walls.walls import get_all_wall_types_by_category
from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm
from duHast.Utilities.utility import encode_ascii

from Autodesk.Revit.DB import Element


def get_wall_report_data(doc, revit_file_path):
    """
    Gets wall data to be written to report file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The file hostname, which is added to data returned.
    :type revit_file_path: str
    :return: list of list of sheet properties.
    :rtype: list of list of str
    """

    data = []
    wall_types = get_all_wall_types_by_category(doc)
    for wt in wall_types:
        try:
            wall_type_name = str(Element.Name.GetValue(wt))
            cs = wt.GetCompoundStructure()
            if cs:
                cs_layers = cs.GetLayers()
                # print(len(cs_layers))
                for cs_layer in cs_layers:
                    layer_mat = doc.GetElement(cs_layer.MaterialId)
                    material_mark  = "N/A"
                    material_name = "N/A"
                    # not all layers may have assigned a material (could be Default)
                    if layer_mat is not None:
                        # get the material mark
                        material_mark = get_element_mark(layer_mat)

                        # convert the default none to n/a
                        if material_mark == "None":
                            material_mark = "N/A"
                        
                        # get the material name
                        material_name = get_material_name_by_id(
                            doc, cs_layer.MaterialId
                        )
                    layer_function = str(cs_layer.Function)
                    layer_width = str(
                        convert_imperial_feet_to_metric_mm(cs_layer.Width)
                    )  # conversion from imperial to metric
                    data.append(
                        [
                            revit_file_path,
                            str(wt.Id),
                            encode_ascii(wall_type_name),
                            layer_function,
                            layer_width,
                            encode_ascii(material_name),
                            encode_ascii(material_mark),
                        ]
                    )
            else:
                data.append(
                    [
                        revit_file_path,
                        str(wt.Id),
                        encode_ascii(wall_type_name),
                        "no layers - in place family or curtain wall",
                        str(0.0),
                        "N/A",
                        "N/A",
                    ]
                )
        except:
            data.append([revit_file_path, str(wt.Id), "N/A", "N/A", "N/A", "N/A", "N/A"])
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families helper functions forcing revit to recognize changes in families and reload it
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This is achieved by creating a dummy type, saving the file, deleting the dummy type and saving the file again

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.family_types import create_family_type, delete_family_type
from duHast.Revit.Common.file_io import save_file

def force_family_update_on_reload(doc, output):
    """
    Forces revit to recognise a change in the family and do a full reload.
    This is achieved by creating a dummy type, saving the file, deleting the dummy type and saving the file again.

    :param doc: Revit Document
    :type doc: Document
    :param output: pyRevit output
    :type output: Output
   

    :return: Result
    """

    # set up a status tracker
    return_value = Result()

    dummy_type_name = "DummyType"

    try:

        # check output, needs to be a function accepting a string
        if not callable(output):
            message = "Output must be a callable function that accepts a string."
            return_value.update_sep(
                False,
                message,
            )
            return return_value

        # check if this is a family document
        if not doc.IsFamilyDocument:
            message = "This is not a family document. Aborting operation."
            output(message)
            return_value.update_sep(
                False,
                message,
            )
            return return_value

        # create a new dummy family type
        create_result = create_family_type(doc, dummy_type_name)
        # check if type was created successfully
        if not create_result.status:
            message =  "Failed to create a dummy family type. Aborting operation."
            output(message)
            return_value.update_sep(
                False,
                message,
            )
            return return_value
        else:
            message = "Successfully created a dummy family type."
            output(message)
            return_value.append_message(message)
        
        # save the family
        save_result = save_file(doc, True)
        # check if family was saved successfully
        if not save_result.status:
            message = "Failed to save the family. Aborting operation."
            output(message)
            return_value.update_sep(
                False,
                message,
            )
            return return_value
        else:
            message = "Successfully saved the family."
            output(message)
            return_value.append_message(message)
        
        # delete the family type
        delete_result = delete_family_type(doc, dummy_type_name)
        # check if type was deleted successfully
        if not delete_result.status:
            message = "Failed to delete the dummy family type. Aborting operation."
            output(message)
            return_value.update_sep(
                False,
                message,
            )
            return return_value
        else:
            message = "Successfully deleted the dummy family type."
            output(message)
            return_value.append_message(message)
       
        # save the family again
        save_again_result = save_file(doc, True)
        # check if family was saved successfully
        if not save_again_result.status:
            message = "Failed to save the family. Aborting operation."
            output(message)
            return_value.update_sep(
                False,
                message,
            )
            return return_value
        else:
            message = "Successfully saved the family."
            output(message)
            return_value.append_message(message)
        
        message = "Successfully forced update to family."
        output(message)
        return_value.append_message(message)

    except Exception as e:
        output("An exception occurred: {}".format(e))
        return_value.update_sep(
            False, "Failed to force update to family with exception: {}".format(e)
        )

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families helper functions retrieving elements from a family.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import (
    CurveElement,
    Element,
    FamilyElementVisibility,
    FamilyElementVisibilityType,
    FilteredElementCollector,
    GenericForm,
    ModelText,
    ReferencePlane,
    Transaction,
)

LINE_NAMES = [
    "Model Lines",  # 3D families
    "Symbolic Lines",  # 3D families
    "Line",  # annotation (tag) families
]


def get_all_generic_forms_in_family(doc):
    """
    Filters all generic forms (3D extrusions) in family.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector of Autodesk.Revit.DB.GenericForm.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    col = FilteredElementCollector(doc).OfClass(GenericForm)
    return col


def get_all_curve_based_elements_in_family(doc):
    """
    Filters all curve-based elements in a family, including symbolic lines and model lines.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of Autodesk.Revit.DB.CurveElement.
    :rtype: list of Autodesk.Revit.DB.CurveElement
    """

    elements = []
    col = FilteredElementCollector(doc).OfClass(CurveElement)
    for c in col:
        if Element.Name.GetValue(c) in LINE_NAMES:
            elements.append(c)
    return elements


def get_all_model_text_elements_in_family(doc):
    """
    Filters all model text elements in family.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector of Autodesk.Revit.DB.ModelText.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    col = FilteredElementCollector(doc).OfClass(ModelText)
    return col


def get_all_reference_planes_in_family(doc):
    """
    Filters all reference planes in family.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector of Autodesk.Revit.DB.ReferencePlane.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    col = FilteredElementCollector(doc).OfClass(ReferencePlane)
    return col


def set_element_visibility_by_detail_level(
        doc, 
        element, 
        detail_level_coarse = True,  
        detail_level_medium = True, 
        detail_level_fine = True, 
        family_element_visibility_type =  FamilyElementVisibilityType.Model, 
        transaction_manager= in_transaction
    ):
    
    """
    Set the visibility of an element by detail level.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: The element to set the visibility for.
    :type element: Autodesk.Revit.DB.Element
    :param detail_level_coarse: Set visibility for coarse detail level.
    :type detail_level_coarse: bool
    :param detail_level_fine: Set visibility for fine detail level.
    :type detail_level_fine: bool
    :param detail_level_medium: Set visibility for medium detail level.
    :type detail_level_medium: bool
    :param family_element_visibility_type: The type of visibility to set. (Model vs View specific)
    :type family_element_visibility_type: Autodesk.Revit.DB.FamilyElementVisibilityType
    
    :return: Result class instance.

        - `result.status` (bool): True if the visibility was updated successfully, otherwise False.
        - `result.message` (str): Confirmation of successful application of visibility settings.
        - `result.result` (list): Empty.
        
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    
    try:
        
        def action():
            action_return_value = Result()
            try:
                fam_element_visibility = FamilyElementVisibility(family_element_visibility_type)

                # Set visibility options
                fam_element_visibility.IsShownInCoarse = detail_level_coarse
                fam_element_visibility.IsShownInMedium = detail_level_medium   
                fam_element_visibility.IsShownInFine = detail_level_fine

                # Apply visibility settings to the extrusion
                element.SetVisibility(fam_element_visibility)

                action_return_value.append_message("Visibility set by detail level fine: {} medium: {} coarse: {}".format(detail_level_fine, detail_level_medium, detail_level_coarse))
            except Exception as e:
                action_return_value.update_sep(False, "Error setting visibility by detail level: {}".format(e))
            return action_return_value
        
        # check if this need to be run inside a transaction
        if transaction_manager:
            # If a transaction manager is provided, create a new transaction
            transaction = Transaction(doc, "Setting detail level visibility")
            return_value = transaction_manager(transaction,action )
        else:
            # If no transaction manager is provided, run the action without a transaction
            # assuming there is one already in place
            return_value = action()
        
    except Exception as e:
        return_value.update_sep(False, "Error setting visibility by detail level: {}".format(e))
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing reporting family type reporting functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reports:

- all family types in the library folder ( based on xml files located in the library folder)

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.family_types_get_data_from_xml import (
    get_family_type_data_from_library,
)
from duHast.Utilities.files_xml import get_all_xml_files_from_directories
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects.result import Result
from duHast.UI.Objects.ProgressBase import ProgressBase


def build_report(type_data_storage_manager_instances):
    """
    Build a report of the family type data retrieved from the library.

    :param type_data_storage_manager_instances: list of matched family type data
    :type type_data_storage_manager_instances: [:class:`FamilyTypeDataStorageManager`]

    :return:
        Result class instance.

        - result.status: Comparison status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message: will be empty.
        - result.result will be [[str]] where each entry is a list of family name, category etc and difference.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    report_data = []
    return_value = Result()
    try:
        # loop over the data and built the report
        for type_data_storage_manager_instance in type_data_storage_manager_instances:
            fam_data = type_data_storage_manager_instance.get_report_data()
            for fam in fam_data:
                report_data.append(fam)
        # store the report data
        return_value.result = report_data
    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )

    return return_value


def get_family_type_data_from_library_xml(process_directories, progress_callback=None):
    """
    Reads family xml atom files from library directories and saves them out to a combined csv report.

    :param process_directories: list of directories to search for xml files
    :type process_directories: list
    :param progress_callback: progress callback object
    :type progress_callback: :class:`ProgressBase`

    :return:
        Result class instance.

        - result.status: Report status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message: Log entries.
        - result.result will be [[str]] where each entry is a list of family name, category etc and difference.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )

    # set up a timer
    t = Timer()
    t.start()

    try:

        # get all xml files from the directory representing families in the library (point of truth)
        xml_files_in_libraries = get_all_xml_files_from_directories(process_directories)

        # check if any xml files were found
        if len(xml_files_in_libraries) == 0:
            return_value.update_sep(
                False,
                "No XML files found in the directories: {}".format(process_directories),
            )
            return_value.append_message(t.stop())
            return return_value
        else:
            return_value.append_message(
                "Found {} XML files in the directories: {}".format(
                    len(xml_files_in_libraries), process_directories
                )
            )

        # get the type data from the library
        type_data_from_library_result = get_family_type_data_from_library(
            xml_files_in_libraries, progress_callback
        )

        # check if the type data from the library was successfully gathered
        if (
            type_data_from_library_result.status == False
            or len(type_data_from_library_result.result) == 0
        ):
            return_value.update_sep(False, type_data_from_library_result.message)
            return_value.append_message(t.stop())
            return return_value
        else:
            return_value.append_message(
                "Successfully gathered family type data from the library."
            )

        # build the report
        report_result = build_report(type_data_from_library_result.result)

        # check if the report was successfully built
        if report_result.status == False:
            return_value.update_sep(False, report_result.message)
            return_value.append_message(t.stop())
            return return_value
        else:
            return_value.append_message(
                "Successfully built the report. {}".format(t.stop())
            )

        # store the report data
        return_value.result = report_result.result

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )
        return_value.append_message(t.stop())

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around data storage in Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Data storage can be used to project wide settings not associated with a specific element. 

Obsolete in Revit 2025 ... 
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import duHast.Utilities.Objects.result as res

from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.ExtensibleSchemas.extensible_schemas import get_schema


from Autodesk.Revit.DB import FilteredElementCollector, Transaction
from Autodesk.Revit.DB.ExtensibleStorage import DataStorage



def find_data_storage(doc, schema_guid):
    """
    Find a DataStorage element in the Revit document that contains the specified schema.
    
    :param doc: The Revit document to search in.
    :type doc: Autodesk.Revit.DB.Document
    :param schema_guid: The GUID of the schema to look for.
    :type schema_guid: str
    
    :return: The DataStorage element if found, None otherwise.
    :rtype: Autodesk.Revit.DB.DataStorage or None
    """
    
    # Look up the schema using its GUID
    schema = get_schema(schema_guid)
    if not schema:
        return None

    # Use a FilteredElementCollector to find all DataStorage elements
    collector = FilteredElementCollector(doc).OfClass(DataStorage)
    for data_storage in collector:
        # Check if the DataStorage element has an entity with the schema
        entity = data_storage.GetEntity(schema)
        if entity and entity.IsValid():
            return data_storage

    return None


def update_entity_on_data_storage(doc, data_storage, entity):
    """
    Update the entity of a DataStorage element with new data.
    
    :param doc: The Revit document to update the DataStorage element in.
    :type doc: Autodesk.Revit.DB.Document
    
    :param data_storage: The DataStorage element to update.
    :type data_storage: Autodesk.Revit.DB.DataStorage
    
    :param entity: The new entity to set on the DataStorage element.
    :type entity: Autodesk.Revit.DB.ExtensibleStorage.Entity

    :return: A Result object containing the updated DataStorage element in its result list or an error message.
    :rtype: duHast.Utilities.Objects.result.Result
    """
    
    return_value = res.Result()
    
    # build an action to update the DataStorage element
    def action():
        action_return_value = res.Result()
        try:
            # Update the DataStorage element with the new Entity
            data_storage.SetEntity(entity)
            action_return_value.append_message(
                "DataStorage element updated successfully."
            )
            action_return_value.result.append(data_storage)
        except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to update data storage: {}".format(e),
                )
        return action_return_value
    
    # attempt to update the DataStorage element in a transaction
    transaction = Transaction(doc, "Updating data storage")
    data_storage_update_result = in_transaction(transaction, action)
    return_value.update(data_storage_update_result)

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Category override to data storage class helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    Element,
    ElementId,
    OverrideGraphicSettings,
    LinePatternElement,
)

from duHast.Revit.Views.Objects.view_graphics_settings import ViewGraphicsSettings
from duHast.Revit.LinePattern.line_patterns import get_all_line_patterns
from duHast.Revit.LinePattern.fill_patterns import get_all_fill_pattern
from duHast.Revit.Categories.categories_model import get_categories_in_model
from duHast.Revit.Views.Objects.Data.override_by_category import OverrideByCategory
from duHast.Revit.Views.Objects.Data.override_by_filter import OverrideByFilter
from duHast.Revit.Views.Objects.Data.override_projection import OverrideProjection
from duHast.Revit.Views.Objects.Data.override_cut import OverrideCut
from duHast.Revit.Common.Objects.Data.pattern_background import PatternBackground
from duHast.Revit.Common.Objects.Data.pattern_foreground import PatternForeground
from duHast.Revit.Common.Objects.Data.line_projection import LineProjection
from duHast.Revit.Common.Objects.Data.line_cut import LineCut
from duHast.Revit.LinePattern.Data.Objects.line_pattern_settings import (
    LinePatternSettings,
)
from duHast.Revit.LinePattern.Data.Objects.fill_pattern_settings import (
    FillPatternSettings,
)
from duHast.Revit.Common.Objects.Data.pattern_settings_base import PatternSettingBase
from duHast.Revit.Common.Utility.revit_to_data_conversion import (
    to_colour,
    VIEW_DETAIL_LEVEL_NAME_MAPPING,
)

from duHast.UI.Objects.ProgressBase import ProgressBase


# ----------------------  utility ----------------------


def _get_name_from_pattern_id(id, pattern_list):
    """
    Returns the matching name of a pattern from the list past in by the id

    :param id: The pattern id to match
    :type id: Autodesk.Revit.DB.ElementId
    :param pattern_list: A list of pattern elements
    :type pattern_list: [PatternElement] (Can be AutoDesk.Revit.LinePatternElement or Autodesk.Revit.DB.FillPatternElement)
    :return: The pattern name if a match was found, "No pattern override assigned." when no pattern was assigned otherwise None (should nt really happen!)
    :rtype: str or None
    """
    if id == LinePatternElement.GetSolidPatternId():
        return PatternSettingBase.SOLID_PATTERN
    elif id == ElementId.InvalidElementId:
        return PatternSettingBase.NO_PATTERN
    for pat in pattern_list:
        if id == pat.Id:
            return Element.Name.GetValue(pat)
    return None


# ----------------------  overrides ----------------------


def get_projection_overrides(view_override_graphic, line_patterns, fill_patterns):
    """
    Retrieves projection override settings from an OverridesGraphicSettings object

    :param view_override_graphic: An overrides graphic setting
    :type view_override_graphic: Autodesk.Revit.DB.OverridesGraphicSettings
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param fill_patterns: All fill patterns in the model
    :type fill_patterns: [Autodesk.Revit.DB.FillPatternElement]

    :return: An overrides projections object instance.
    :rtype: :class:`.OverrideProjection`
    """

    # set up data storage class instance
    projection_override = OverrideProjection()

    # start
    # projection line pattern
    line_pat = LineProjection()
    line_pat.colour = to_colour(view_override_graphic.ProjectionLineColor)
    line_pat.weight = view_override_graphic.ProjectionLineWeight
    # basic line pattern details
    line_pat_settings = LinePatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.ProjectionLinePatternId, pattern_list=line_patterns
        ),
        id=view_override_graphic.ProjectionLinePatternId.IntegerValue,
    )
    # store line pattern details
    line_pat.line_pattern_settings = line_pat_settings
    # store projection line pattern details
    projection_override.line_projection = line_pat

    # projection pattern foreground
    fore_pat = PatternForeground()
    fore_pat.colour = to_colour(view_override_graphic.SurfaceForegroundPatternColor)
    fore_pat.is_visible = view_override_graphic.IsSurfaceForegroundPatternVisible
    # basic foreground pattern details
    fore_pat_settings = FillPatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.SurfaceForegroundPatternId,
            pattern_list=fill_patterns,
        ),
        id=view_override_graphic.SurfaceForegroundPatternId.IntegerValue,
    )
    # store foreground pattern settings details
    fore_pat.fill_pattern_setting = fore_pat_settings
    # store foreground pattern in override
    projection_override.pattern_foreground = fore_pat

    # projection pattern background
    back_pat = PatternBackground()
    back_pat.colour = to_colour(view_override_graphic.SurfaceBackgroundPatternColor)
    back_pat.is_visible = view_override_graphic.IsSurfaceBackgroundPatternVisible
    # basic background pattern details
    back_pat_settings = FillPatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.SurfaceBackgroundPatternId,
            pattern_list=fill_patterns,
        ),
        id=view_override_graphic.SurfaceBackgroundPatternId.IntegerValue,
    )
    # store foreground pattern settings details
    back_pat.fill_pattern_setting = back_pat_settings
    # store background pattern in override
    projection_override.pattern_background = back_pat

    return projection_override


def get_cut_overrides(view_override_graphic, line_patterns, fill_patterns):
    """
    Retrieves cut override settings from an OverridesGraphicSettings object

    :param view_override_graphic: An overrides graphic setting
    :type view_override_graphic: Autodesk.Revit.DB.OverridesGraphicSettings
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param fill_patterns: All fill patterns in the model
    :type fill_patterns: [Autodesk.Revit.DB.FillPatternElement]

    :return: An overrides projections object instance.
    :rtype: :class:`.OverrideCut`
    """

    # set up data storage class instance
    cut_override = OverrideCut()

    # start
    # cut line pattern
    line_pat = LineCut()
    line_pat.colour = to_colour(view_override_graphic.CutLineColor)
    line_pat.weight = view_override_graphic.CutLineWeight
    # basic line pattern details
    line_pat_settings = LinePatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.CutLinePatternId, pattern_list=line_patterns
        ),
        id=view_override_graphic.CutLinePatternId.IntegerValue,
    )
    # store line pattern details
    line_pat.line_pattern_settings = line_pat_settings
    # store projection line pattern details
    cut_override.line_cut = line_pat

    # cut pattern foreground
    fore_pat = PatternForeground()
    fore_pat.colour = to_colour(view_override_graphic.CutForegroundPatternColor)
    fore_pat.is_visible = view_override_graphic.IsCutForegroundPatternVisible
    # basic foreground pattern details
    fore_pat_settings = FillPatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.CutForegroundPatternId, pattern_list=fill_patterns
        ),
        id=view_override_graphic.CutForegroundPatternId.IntegerValue,
    )
    # store foreground pattern settings details
    fore_pat.fill_pattern_setting = fore_pat_settings
    # store foreground pattern in override
    cut_override.pattern_foreground = fore_pat

    # cut pattern background
    back_pat = PatternBackground()
    back_pat.colour = to_colour(view_override_graphic.CutBackgroundPatternColor)
    back_pat.is_visible = view_override_graphic.IsCutBackgroundPatternVisible
    # basic background pattern details
    back_pat_settings = FillPatternSettings(
        name=_get_name_from_pattern_id(
            id=view_override_graphic.CutBackgroundPatternId, pattern_list=fill_patterns
        ),
        id=view_override_graphic.CutBackgroundPatternId.IntegerValue,
    )
    # store foreground pattern settings details
    back_pat.fill_pattern_setting = back_pat_settings
    # store background pattern in override
    cut_override.pattern_background = back_pat

    return cut_override


# ----------------------  by category overrides ----------------------


def get_view_category_overrides(
    view, categories_in_model, line_patterns, fill_patterns
):
    """
    Returns the category overrides for all categories past in for a given view.

    :param view: The view from which to retrieve the category overrides.
    :type view: Autodesk.Revit.DB.View
    :param categories_in_model: All categories in the model.
    :type categories_in_model: _type_
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param fill_patterns: All fill patterns in the model
    :type fill_patterns: [Autodesk.Revit.DB.FillPatternElement]

    :return: A list of all category overrides for a view.
    :rtype: [:class:`.OverrideByCategory`]
    """

    overrides_data = []
    for model_cat in categories_in_model:
        # set up storage class
        override = OverrideByCategory(
            main_category_name=model_cat.category_name,
            sub_category_name=model_cat.sub_category_name,
            category_id=model_cat.id.IntegerValue,
        )

        # get the overrides by category
        view_override = view.GetCategoryOverrides(model_cat.id)

        # get general overrides settings
        override.halftone = view_override.Halftone
        override.transparency = view_override.Transparency

        # get cut overrides for category
        override.override_cut = get_cut_overrides(
            view_override_graphic=view_override,
            line_patterns=line_patterns,
            fill_patterns=fill_patterns,
        )

        # get projection for category
        override.override_projection = get_projection_overrides(
            view_override_graphic=view_override,
            line_patterns=line_patterns,
            fill_patterns=fill_patterns,
        )

        # check if category is hidden ( need to inverse since I'm storing ( is visible) the exact opposite (is hidden) of what is returned)
        override.is_visible = not (view.GetCategoryHidden(model_cat.id))

        # get the detail level as an integer
        override.detail_level = VIEW_DETAIL_LEVEL_NAME_MAPPING[
            view_override.DetailLevel
        ]

        # check if any override has been applied by comparing the retrieved override
        # with a default one
        default_override = OverrideByCategory()
        if default_override.compare_overrides(override):
            override.are_overrides_present = False
        else:
            override.are_overrides_present = True

        # save overrides in list to be returned
        overrides_data.append(override)

    return overrides_data


# ----------------------  by filter overrides ----------------------


def get_view_filter_overrides(doc, view, line_patterns, fill_patterns):
    """
    Returns the filter overrides for all filter applied to a given view.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view from which to retrieve the category overrides.
    :type view: Autodesk.Revit.DB.View
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param fill_patterns: All fill patterns in the model
    :type fill_patterns: [Autodesk.Revit.DB.FillPatternElement]

    :return: A list of all filter overrides for a view.
    :rtype: [:class:`.OverrideByFilter`]
    """

    overrides_data = []

    # get all filters applied to a view
    filter_ids_applied = view.GetFilters()

    for filter_id in filter_ids_applied:
        # get the actual filter element
        filter = doc.GetElement(filter_id)
        # set up storage class
        override = OverrideByFilter(
            filter_name=Element.Name.GetValue(filter),
            filter_id=filter_id.IntegerValue,
        )

        # get the overrides by filter
        view_override = view.GetFilterOverrides(filter_id)
        # get general overrides settings
        override.halftone = view_override.Halftone
        override.transparency = view_override.Transparency

        # get cut overrides for filter
        override.override_cut = get_cut_overrides(
            view_override_graphic=view_override,
            line_patterns=line_patterns,
            fill_patterns=fill_patterns,
        )

        # get projection for filter
        override.override_projection = get_projection_overrides(
            view_override_graphic=view_override,
            line_patterns=line_patterns,
            fill_patterns=fill_patterns,
        )

        # check if filter switches things off
        override.is_visible = view.GetFilterVisibility(filter_id)

        # check if filter is enabled
        override.is_enabled = view.GetIsFilterEnabled(filter_id)

        # check if any override has been applied by comparing the retrieved override
        # with a default one
        default_override = OverrideByFilter()
        if default_override.compare_overrides(override):
            override.are_overrides_present = False
        else:
            override.are_overrides_present = True

        # save overrides in list to be returned
        overrides_data.append(override)

    return overrides_data


def get_view_settings(doc, view):
    """
    retrieves the filter and category graphic overrides from a view

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: _description_
    :type view: _type_
    :return: A ViewGraphicsSettings instance with retrieved setting.
    :rtype: :class:`.ViewGraphicsSettings`
    """

    # set up class to store view graphic settings
    view_data = ViewGraphicsSettings(
        view_name=Element.Name.GetValue(view),
        view_id=view.Id.IntegerValue,
    )

    # get all pattern element since they will be required in the override settings
    # get all line pattern and fill patterns in model
    line_pattern_elements = get_all_line_patterns(doc=doc)

    # get all fill pattern elements in model
    fill_pattern_elements = get_all_fill_pattern(doc=doc)

    # retrieve all categories available in model
    categories_in_model = get_categories_in_model(doc=doc)

    # get category graphic overrides
    category_overrides = get_view_category_overrides(
        view=view,
        categories_in_model=categories_in_model,
        line_patterns=line_pattern_elements,
        fill_patterns=fill_pattern_elements,
    )
    view_data.override_by_category = category_overrides

    # get filter graphic overrides
    filter_overrides = get_view_filter_overrides(
        doc=doc,
        view=view,
        line_patterns=line_pattern_elements,
        fill_patterns=fill_pattern_elements,
    )
    view_data.override_by_filter = filter_overrides

    return view_data


def get_views_graphic_settings_data(doc, views, progress_callback=None):
    """
    Gets view data graphic settings from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param views: Views of which to report graphical overrides on. (View must support graphical overrides, otherwise an exception will be thrown!)
    :type views: [Autodesk.Revit.DB.View]
    :progress_callback: A progress call back object, default is None
    :type progress_callback: :class:`.ProgressBase` or None
    :return: list of ViewGraphicsSettings instances
    :rtype: [:class:`.ViewGraphicsSettings`]
    """

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )
    
    if (isinstance(views, list) == False):
        raise TypeError("views must be a list of Autodesk.Revit.DB.View instances.")

    # set up list to store view settings
    views_settings = []

    # check if there are any views to process
    if len(views) == 0:
        return views_settings
    
    # sort views by name
    views = sorted(views, key=lambda x: Element.Name.GetValue(x))

    # progress call back
    callback_counter = 1

    # loop over past in views and retrieve settings
    for view in views:

        if progress_callback != None:
            progress_callback.update(callback_counter, len(views), "{}".format(Element.Name.GetValue(view)))

        view_setting = get_view_settings(
            doc=doc,
            view=view,
        )
        views_settings.append(view_setting)

        # increment counter
        callback_counter += 1
        # check for user cancel
        if progress_callback != None:
            if progress_callback.is_cancelled():
                break

    return views_settings
